!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                                   !!
!!   Copyright 2012 David Alonso                                     !!
!!                                                                   !!
!!                                                                   !!
!! This file is part of CRIME.                                       !!
!!                                                                   !!
!! CRIME is free software: you can redistribute it and!or modify it  !!
!! under the terms of the GNU General Public License as published by !!
!! the Free Software Foundation, either version 3 of the License, or !!
!! (at your option) any later version.                               !!
!!                                                                   !!
!! CRIME is distributed in the hope that it will be useful, but      !!
!! WITHOUT ANY WARRANTY; without even the implied warranty of        !!
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU !!
!! General Public License for more details.                          !!
!!                                                                   !!
!! You should have received a copy of the GNU General Public License !!
!! along with CRIME.  If not, see <http://www.gnu.org/licenses/>.    !!
!!                                                                   !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM ForGet
  USE omp_lib
  USE params_type
  USE io4G
  USE frequencies
  USE galaxy
  USE angles
  USE polarization

  implicit none
  CHARACTER(len=256) :: fname_input
  type(Params4G) :: pars

  IF(COMMAND_ARGUMENT_COUNT().ne.1) THEN
     WRITE(*,*) "Usage: ForGet param_filename"
     CALL exit(1)
  END IF
  CALL get_command_argument(1,fname_input)

  CALL read_input_params_ForGet(fname_input,pars)

  IF(pars%do_galaxy==1) THEN
     CALL ForGet_galactic(pars)
  ELSE
     CALL ForGet_homogeneous(pars)
  END IF

  CALL free_params_ForGet(pars)

CONTAINS

  SUBROUTINE ForGet_homogeneous(pars)
    implicit none
    type(Params4G), intent(in) :: pars
    REAL(8), allocatable, dimension(:) :: eigenvalues
    REAL(8), allocatable, dimension(:,:) :: eigenvectors
    REAL(FPP), allocatable, dimension(:,:) :: skymap
    COMPLEX(FPP), allocatable, dimension(:,:,:,:) :: alms
    COMPLEX(FPP), allocatable, dimension(:) :: vec_aux
    INTEGER(4) :: stat_alloc, ii, ll, mm
    CHARACTER(len=80), dimension(60) :: header
    CHARACTER(len=256) :: fname_out
    CHARACTER(len=32) :: buffer
    INTEGER(4) :: seed_thr,pid,seed0
    type(planck_rng) :: rng

    CALL set_angles(pars)

    ALLOCATE(vec_aux(1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","vec_aux")
    ALLOCATE(eigenvalues(1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","eigenvalues")
    ALLOCATE(eigenvectors(1:pars%n_nu,1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","eigenvectors")
    ALLOCATE(alms(1:pars%n_nu,1:1,0:pars%lmax,0:pars%lmax),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","alms")
    ALLOCATE(skymap(0:nside2npix(pars%nside)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","skymap")

    CALL decorrelate_frequencies(pars,eigenvectors,eigenvalues)
    
    WRITE(*,*) "ForGet: generating uncorrelated alms"
!$omp parallel default(none), &
!$omp private(ii,seed0,seed_thr,pid,rng), &
!$omp shared(pars,eigenvalues,alms)
    seed0=pars%seed
    pid=omp_get_thread_num()
    seed_thr=seed0+pid
    CALL rand_init(rng,seed_thr)
!$omp do
    DO ii=1,pars%n_nu
       CALL get_alms(rng,pars%lmax,eigenvalues(ii),alms(ii,:,:,:))
    END DO
!$omp end do
!$omp end parallel

    WRITE(*,*) "ForGet: correlating alms"
    DO ll=0,pars%lmax
       DO mm=0,ll
          vec_aux=matmul(eigenvectors,alms(:,1,ll,mm))
          alms(:,1,ll,mm)=vec_aux
       END DO
    END DO
    
    WRITE(*,*) "ForGet: writing skymaps"
    DO ii=1,pars%n_nu
       CALL alm2map(pars%nside,pars%lmax,pars%lmax,alms(ii,:,:,:),skymap(:,1))
       CALL get_map_header_title(pars,header)
       WRITE(buffer,'(I3.3)') ii
       fname_out=TRIM(pars%prefix_out)//"_"//TRIM(buffer)//".fits"
       CALL write_map(fname_out,skymap,header)
    END DO

    CALL end_angles()
    DEALLOCATE(eigenvalues,eigenvectors,alms,vec_aux,skymap)
  END SUBROUTINE ForGet_homogeneous
  
  SUBROUTINE ForGet_galactic(pars)
    implicit none
    type(Params4G), intent(in) :: pars
    REAL(8), allocatable, dimension(:) :: eigenvalues
    REAL(8), allocatable, dimension(:,:) :: eigenvectors
    REAL(FPP), allocatable, dimension(:,:) :: skymaps
    REAL(FPP), allocatable, dimension(:,:) :: skymap_dum
    REAL(FPP), allocatable, dimension(:,:) :: skymaps_pol_r
    REAL(FPP), allocatable, dimension(:,:) :: skymaps_pol_i
    COMPLEX(FPP), allocatable, dimension(:,:,:,:) :: alms
    COMPLEX(FPP), allocatable, dimension(:) :: vec_aux
    INTEGER(4) :: stat_alloc, ii, ll, mm
    CHARACTER(len=80), dimension(60) :: header
    CHARACTER(len=256) :: fname_out
    CHARACTER(len=32) :: buffer
    INTEGER(4) :: seed_thr,pid,seed0
    type(planck_rng) :: rng

    CALL init_galaxy_template(pars)
    CALL set_angles(pars)

    ALLOCATE(vec_aux(1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","vec_aux")
    ALLOCATE(eigenvalues(1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","eigenvalues")
    ALLOCATE(eigenvectors(1:pars%n_nu,1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","eigenvectors")
    ALLOCATE(alms(1:pars%n_nu,1:1,0:pars%lmax,0:pars%lmax),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","alms")
    ALLOCATE(skymaps(0:nside2npix(pars%nside)-1,1:pars%n_nu),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"ForGet","skymaps")

    CALL decorrelate_frequencies(pars,eigenvectors,eigenvalues)

    WRITE(*,*) "ForGet: generating uncorrelated alms"
!$omp parallel default(none), &
!$omp private(ii,seed0,seed_thr,pid,rng), &
!$omp shared(pars,eigenvalues,alms)
    seed0=pars%seed
    pid=omp_get_thread_num()
    seed_thr=seed0+pid
    CALL rand_init(rng,seed_thr)
!$omp do
    DO ii=1,pars%n_nu
       CALL get_alms(rng,pars%lmax,eigenvalues(ii),alms(ii,:,:,:))
    END DO
!$omp end do
!$omp end parallel
    CALL constrain_alms(pars%n_nu,eigenvectors,alms)

    WRITE(*,*) "ForGet: correlating alms"
    DO ll=0,pars%lmax
       DO mm=0,ll
          vec_aux=matmul(eigenvectors,alms(:,1,ll,mm))
          alms(:,1,ll,mm)=vec_aux
       END DO
    END DO

    WRITE(*,*) "ForGet: adding Haslam map"
!$omp parallel default(none), &
!$omp private(ii), shared(pars,alms,skymaps)
!$omp do
    DO ii=1,pars%n_nu
       CALL alm2map(pars%nside,pars%lmax,pars%lmax,alms(ii,:,:,:),skymaps(:,ii))
       CALL add_haslam(pars,ii,skymaps(:,ii))
    END DO
!$omp end do
!$omp end parallel
    DEALLOCATE(eigenvalues,eigenvectors,alms,vec_aux)
    
    IF(pars%do_polarization==1) THEN
       WRITE(*,*) "ForGet: calculating polarization"
       ALLOCATE(skymaps_pol_r(0:nside2npix(NSIDE_POL)-1,1:pars%n_nu),STAT=stat_alloc)
       CALL assert_alloc(stat_alloc,"ForGet","skymaps_pol_r")
       ALLOCATE(skymaps_pol_i(0:nside2npix(NSIDE_POL)-1,1:pars%n_nu),STAT=stat_alloc)
       CALL assert_alloc(stat_alloc,"ForGet","skymaps_pol_i")
       ALLOCATE(skymap_dum(0:nside2npix(pars%nside)-1,1:3),STAT=stat_alloc)
       CALL assert_alloc(stat_alloc,"ForGet","skymap_dum")
       CALL do_polarization(pars,skymaps_pol_r,skymaps_pol_i)
    END IF

    WRITE(*,*) "ForGet: writing skymaps"
    DO ii=1,pars%n_nu
       CALL get_map_header_title(pars,header)
       WRITE(buffer,'(I3.3)') ii
       fname_out=TRIM(pars%prefix_out)//"_"//TRIM(buffer)//".fits"

       IF(pars%do_polarization==1) THEN
          skymap_dum(:,1)=skymaps(:,ii)
          CALL udgrade_ring(skymaps_pol_r(:,ii),NSIDE_POL,skymap_dum(:,2),pars%nside)
          CALL udgrade_ring(skymaps_pol_i(:,ii),NSIDE_POL,skymap_dum(:,3),pars%nside)
          CALL write_map(fname_out,skymap_dum,header)
       ELSE
          CALL write_map(fname_out,skymaps(:,ii:ii),header)
       END IF
    END DO

    CALL end_angles()
    CALL end_galaxy_template()
    DEALLOCATE(skymaps)
    IF(pars%do_polarization==1) THEN
       DEALLOCATE(skymaps_pol_r,skymaps_pol_i,skymap_dum)
    END IF
  END SUBROUTINE ForGet_galactic
  
  SUBROUTINE add_haslam(pars,num_map,skymap)
    implicit none
    type(Params4G), intent(in) :: pars
    INTEGER(4), intent(in) :: num_map
    REAL(FPP), dimension(0:), intent(out) :: skymap
    REAL(8) :: amplitude,t_haslam,specin,nu
    INTEGER(8) :: ii

    nu=pars%nutable(num_map,1)
    WRITE(*,*) "add_haslam: adding Haslam extrapolated to nu = ",nu
    DO ii=0,nside2npix(pars%nside)-1
       CALL get_galaxy_info(ii,t_haslam,specin,amplitude)
       skymap(ii)=t_haslam*(nu/NU_HASLAM)**specin+amplitude*skymap(ii)
    END DO
  END SUBROUTINE add_haslam

  SUBROUTINE get_map_header_title(pars,header)
    implicit none
    type(Params4G), intent(in) :: pars
    CHARACTER(len=80), dimension(:), intent(out) :: header

    IF(pars%do_polarization==1) THEN
       CALL write_minimal_header(header,'MAP',nside=pars%nside,ordering='RING',&
            coordsys='G',creator='ForGet',randseed=pars%seed,polar=.TRUE.)
    ELSE
       CALL write_minimal_header(header,'MAP',nside=pars%nside,ordering='RING',&
            coordsys='G',creator='ForGet',randseed=pars%seed,polar=.FALSE.)
    END IF
  END SUBROUTINE get_map_header_title
  
END PROGRAM ForGet
