!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                                   !!
!!   Copyright 2012 David Alonso                                     !!
!!                                                                   !!
!!                                                                   !!
!! This file is part of CRIME.                                       !!
!!                                                                   !!
!! CRIME is free software: you can redistribute it and!or modify it  !!
!! under the terms of the GNU General Public License as published by !!
!! the Free Software Foundation, either version 3 of the License, or !!
!! (at your option) any later version.                               !!
!!                                                                   !!
!! CRIME is distributed in the hope that it will be useful, but      !!
!! WITHOUT ANY WARRANTY; without even the implied warranty of        !!
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU !!
!! General Public License for more details.                          !!
!!                                                                   !!
!! You should have received a copy of the GNU General Public License !!
!! along with CRIME.  If not, see <http://www.gnu.org/licenses/>.    !!
!!                                                                   !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM JoinT

  USE params_type
  USE io4G
  USE misc_utils
  USE fitstools
  USE pix_tools
  USE head_fits
  USE rngmod
  USE alm_tools

  implicit none

  CHARACTER(len=256) :: fname_input, buffer, fname_out
  CHARACTER(len=80), dimension(60) :: header
  type(ParamsJoinT) :: pars
  REAL(FPP), allocatable, dimension(:,:) :: map_merge
  INTEGER(4) :: ii, stat_alloc

  IF(COMMAND_ARGUMENT_COUNT().ne.1) THEN
     WRITE(*,*) "Usage: JoinT param_filename"
     CALL exit(1)
  END IF
  CALL get_command_argument(1,fname_input)
  
  CALL read_input_params_JoinT(fname_input,pars)
  CALL write_minimal_header(header,'MAP',nside=pars%nside,ordering='RING',&
       coordsys='G',creator='GetHI+ForGet+JoinT')
  ALLOCATE(map_merge(0:nside2npix(pars%nside)-1,1:1),STAT=stat_alloc)
  CALL assert_alloc(stat_alloc,"JoinT","map_merge");

  DO ii=1,pars%n_nu
     WRITE(*,*) "JoinT: merging maps for ",ii,"-th catalog"
     WRITE(buffer,'(I3.3)') ii
     fname_out=TRIM(pars%prefix_catalog)//"_"//TRIM(buffer)//".fits"
     CALL merge_maps(pars,map_merge,ii)
     IF(pars%beam_fwhm>0) THEN
        CALL smooth_map(pars,map_merge)
     END IF
     IF(pars%tsys>0) THEN
        CALL add_noise(pars,map_merge,ii)
     END IF
     CALL output_map(map_merge,header,fname_out)
  END DO

  DEALLOCATE(map_merge)
  CALL free_params_JoinT(pars)

CONTAINS

  SUBROUTINE add_noise(pars,map_in,i_map)
    implicit none
    type(ParamsJoinT), intent(in) :: pars
    REAL(FPP), dimension(0:,:), intent(inout) :: map_in
    INTEGER(4), intent(in) :: i_map
    REAL(8), PARAMETER :: PREFAC_TNU=3.6E9 ! (1 h)*(1 MHz)
    INTEGER(4) :: ii
    REAL(8) :: dnu,nu,sigma_noise
    type(planck_rng) :: rng

    CALL rand_init(rng,pars%seed)
    nu=pars%nutable(i_map,1)
    dnu=pars%nutable(i_map,3)-pars%nutable(i_map,2)
    sigma_noise=pars%tsys/sqrt(PREFAC_TNU*dnu*pars%time_int)

    DO ii=0,nside2npix(pars%nside)-1
       map_in(ii,1)=map_in(ii,1)+sigma_noise*rand_gauss(rng)
    END DO
  END SUBROUTINE add_noise

  SUBROUTINE smooth_map(pars,map_in)
    implicit none
    type(ParamsJoinT), intent(in) :: pars
    REAL(FPP), dimension(0:,:), intent(inout) :: map_in
    REAL(8), PARAMETER :: DEG2AMIN=60.
    COMPLEX(FPP), allocatable, dimension(:,:,:) :: alm
    REAL(8), allocatable, dimension(:,:) :: dw8
    INTEGER(4) :: stat_alloc
    REAL(8), dimension(2) :: z
    REAL(FPP) :: fwhm_dum

    fwhm_dum=DEG2AMIN*pars%beam_fwhm
    ALLOCATE(alm(1:1,0:3*pars%nside-1,0:3*pars%nside-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"smooth_map","alm")
    ALLOCATE(dw8(1:2*pars%nside,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"smooth_map","dw8")
    dw8=1
    z(1)=-1
    z(2)=1
    CALL map2alm(pars%nside,3*pars%nside-1,3*pars%nside-1,&
         map_in(:,1),alm,z,dw8)
    CALL alter_alm(pars%nside,3*pars%nside-1,3*pars%nside-1,&
         fwhm_dum,alm)
    CALL alm2map(pars%nside,3*pars%nside-1,3*pars%nside-1,&
         alm,map_in(:,1))
    DEALLOCATE(alm,dw8)
  END SUBROUTINE smooth_map

  SUBROUTINE merge_maps(pars,map_merge,i_map)
    implicit none
    type(ParamsJoinT), intent(in) :: pars
    INTEGER(4), intent(in) :: i_map
    REAL(FPP), dimension(0:,:), intent(out) :: map_merge
    REAL(FPP), allocatable, dimension(:,:) :: map_read
    REAL(FPP), allocatable, dimension(:,:) :: map_pol
    CHARACTER(len=256) :: fname, buffer
    INTEGER(4) :: stat_alloc

    ALLOCATE(map_read(0:nside2npix(pars%nside)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"merge_maps","map_merge");
    WRITE(buffer,'(I3.3)') i_map
    
    map_merge=0.0

    !Synchrotron
    fname=TRIM(pars%prefix_synchrotron)//"_"//TRIM(buffer)//".fits"
    IF(pars%do_polarization==1) THEN
       ALLOCATE(map_pol(0:nside2npix(pars%nside)-1,1:3),STAT=stat_alloc)
       CALL assert_alloc(stat_alloc,"merge_maps","map_pol");
       CALL input_map(fname,map_pol,nside2npix(pars%nside),3)
       map_read=map_pol(:,1:1)
       map_pol(:,1)=pars%polarization_leakage*map_pol(:,2)!sqrt(map_pol(:,2)**2+map_pol(:,3)**2)
       map_merge=map_merge+map_read+map_pol(:,1:1)
       DEALLOCATE(map_pol)
    ELSE
       CALL input_map(fname,map_read,nside2npix(pars%nside),1)
       map_merge=map_merge+map_read
    END IF

    !Point sources
    fname=TRIM(pars%prefix_point_sources)//"_"//TRIM(buffer)//".fits"
    CALL input_map(fname,map_read,nside2npix(pars%nside),1)
    map_merge=map_merge+map_read

    !Galactic free-free
    fname=TRIM(pars%prefix_galactic_freefree)//"_"//TRIM(buffer)//".fits"
    CALL input_map(fname,map_read,nside2npix(pars%nside),1)
    map_merge=map_merge+map_read

    !Extragalactic free-free
    fname=TRIM(pars%prefix_extragalactic_freefree)//"_"//TRIM(buffer)//".fits"
    CALL input_map(fname,map_read,nside2npix(pars%nside),1)
    map_merge=map_merge+map_read

    !Cosmological signal
    fname=TRIM(pars%prefix_cosmo)//"_"//TRIM(buffer)//".fits"
    CALL input_map(fname,map_read,nside2npix(pars%nside),1)
    map_merge=map_merge+map_read
    
    DEALLOCATE(map_read)
  END SUBROUTINE merge_maps

END PROGRAM JoinT
