!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                                   !!
!!   Copyright 2012 David Alonso                                     !!
!!                                                                   !!
!!                                                                   !!
!! This file is part of CRIME.                                       !!
!!                                                                   !!
!! CRIME is free software: you can redistribute it and!or modify it  !!
!! under the terms of the GNU General Public License as published by !!
!! the Free Software Foundation, either version 3 of the License, or !!
!! (at your option) any later version.                               !!
!!                                                                   !!
!! CRIME is distributed in the hope that it will be useful, but      !!
!! WITHOUT ANY WARRANTY; without even the implied warranty of        !!
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU !!
!! General Public License for more details.                          !!
!!                                                                   !!
!! You should have received a copy of the GNU General Public License !!
!! along with CRIME.  If not, see <http://www.gnu.org/licenses/>.    !!
!!                                                                   !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE galaxy

  USE params_type
  USE io4G
  USE alm_tools
  USE pix_tools
  USE fitstools
  USE head_fits

  implicit none

  INTEGER(4), PARAMETER :: NSIDE_TEMPLATE=512
  INTEGER(4), PARAMETER :: NSIDE_REGIONS=16
  INTEGER(4), PARAMETER :: LMAX_READ=1024
  INTEGER(4), PARAMETER :: LMAX_HASLAM=200
  REAL(8), PARAMETER :: NU_HASLAM=408.0
  REAL(8), PARAMETER :: SPECIN_OFFSET=0.0
  REAL(FPP), allocatable, dimension(:,:) :: galaxy_amplitude
  REAL(FPP), allocatable, dimension(:,:) :: galaxy_template
  REAL(FPP), allocatable, dimension(:,:) :: specin_template
  COMPLEX(FPP), allocatable, dimension(:,:,:) :: alm_haslam
  INTEGER(4) :: i_haslam

CONTAINS

  SUBROUTINE test_galaxy_maps()
    implicit none
    CHARACTER(len=80), dimension(60) :: header

    CALL write_minimal_header(header,'map',nside=NSIDE_TEMPLATE,&
         ordering='RING',coordsys='G',creator='test_galaxy_maps')
  
    CALL write_map("haslam_read.fits",galaxy_template,header)
    CALL write_map("specin_read.fits",specin_template,header)
    CALL write_map("amplitude_read.fits",galaxy_amplitude,header)

  END SUBROUTINE test_galaxy_maps

  SUBROUTINE init_galaxy_template(pars)
    implicit none
    type(Params4G), intent(in) :: pars
    REAL(FPP), allocatable, dimension(:) :: map_mean, map_sigma
    INTEGER(4), allocatable, dimension(:) :: n_pixels
    INTEGER(4) :: ii, ipix_hi, ipix_lo, stat_alloc
    REAL(8) :: tval
    REAL(8) :: theta, phi, nu0, nuf
    REAL(8), allocatable, dimension(:,:) :: dw8
    REAL(8), dimension(2) :: z
     
    i_haslam=-1
    DO ii=1,pars%n_nu
       nu0=pars%nutable(ii,2)
       nuf=pars%nutable(ii,3)
       IF((NU_HASLAM<=nuf).and.(NU_HASLAM>nu0)) THEN
          i_haslam=ii
          EXIT
       END IF
    END DO

    IF(i_haslam==-1) THEN
       WRITE(*,*) "init_galaxy_template: The Haslam map is outside the frequency range"
       CALL exit(1)
    END IF
    
    ALLOCATE(galaxy_amplitude(0:nside2npix(NSIDE_TEMPLATE)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","galaxy_amplitude") 
    ALLOCATE(galaxy_template(0:nside2npix(NSIDE_TEMPLATE)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","galaxy_template")
    ALLOCATE(specin_template(0:nside2npix(NSIDE_TEMPLATE)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","specin_template")

    ALLOCATE(map_mean(0:nside2npix(NSIDE_REGIONS)-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","map_mean")
    ALLOCATE(map_sigma(0:nside2npix(NSIDE_REGIONS)-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","map_sigma")
    ALLOCATE(n_pixels(0:nside2npix(NSIDE_REGIONS)-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","n_pixels")

    WRITE(*,*) "init_galaxy_template: Reading Haslam map from file: ",&
         TRIM(pars%fname_haslam)
    CALL input_map(pars%fname_haslam,galaxy_template,nside2npix(NSIDE_TEMPLATE),1)
    CALL convert_inplace(nest2ring,galaxy_template(:,1))
    galaxy_template(:,1)=galaxy_template(:,1)*1000; !Transform from K to mK

    WRITE(*,*) "init_galaxy_template: Reading spectral index map from file: ",&
         TRIM(pars%fname_specin)
    CALL input_map(pars%fname_specin,specin_template,nside2npix(NSIDE_TEMPLATE),1)
    specin_template(:,1)=specin_template(:,1)-2.0+SPECIN_OFFSET

    WRITE(*,*) "init_galaxy_templates: Calculating rms in regions"
    n_pixels=0
    map_mean=0
    map_sigma=0
    DO ii=0,INT(nside2npix(NSIDE_TEMPLATE)-1)
       ipix_hi=ii
       CALL pix2ang_ring(NSIDE_TEMPLATE,ipix_hi,theta,phi)
       CALL ang2pix_ring(NSIDE_REGIONS,theta,phi,ipix_lo)
       tval=galaxy_template(ipix_hi,1)
       n_pixels(ipix_lo)=n_pixels(ipix_lo)+1
       map_mean(ipix_lo)=map_mean(ipix_lo)+tval
       map_sigma(ipix_lo)=map_sigma(ipix_lo)+tval*tval
    END DO
    map_mean=map_mean/n_pixels
    map_sigma=map_sigma/n_pixels-map_mean*map_mean
    map_sigma=map_sigma/MAXVAL(map_sigma)
    map_sigma=sqrt(map_sigma)

    DO ii=0,INT(nside2npix(NSIDE_TEMPLATE)-1)
       ipix_hi=ii
       CALL pix2ang_ring(NSIDE_TEMPLATE,ipix_hi,theta,phi)
       CALL ang2pix_ring(NSIDE_REGIONS,theta,phi,ipix_lo)
       galaxy_amplitude(ipix_hi,1)=map_sigma(ipix_lo)
    END DO

    WRITE(*,*) "init_galaxy_template: removing redundant multipoles"
    ALLOCATE(alm_haslam(1:1,0:LMAX_READ,0:LMAX_READ),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","alm_haslam")
    ALLOCATE(dw8(1:2*NSIDE_TEMPLATE,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_galaxy_template","dw8")
    dw8=1.0
    z(1)=-1
    z(2)=1
    CALL map2alm(NSIDE_TEMPLATE,LMAX_READ,LMAX_READ,&
         galaxy_template(:,1),alm_haslam,z,dw8)
    DO ii=LMAX_HASLAM+1,LMAX_READ
       alm_haslam(1:1,ii,:)=0
    END DO
    CALL alm2map(NSIDE_TEMPLATE,LMAX_READ,LMAX_READ,alm_haslam,&
         galaxy_template(:,1))

    DEALLOCATE(map_mean,map_sigma,n_pixels,dw8)

!    CALL test_galaxy_maps()
  END SUBROUTINE init_galaxy_template

  SUBROUTINE constrain_alms(n_nu,eigenvectors,alms)
    implicit none
    INTEGER(4), intent(in) :: n_nu
    REAL(8), dimension(n_nu,n_nu), intent(in) :: eigenvectors
    COMPLEX(FPP), dimension(:,:,0:,0:), intent(inout) :: alms

    INTEGER(4) :: ll,mm

    WRITE(*,*) "constrain_alms: constraining alms to Haslam map"

    DO ll=0,LMAX_HASLAM
       DO mm=0,ll
          alms(n_nu,1,ll,mm)=DOT_PRODUCT(eigenvectors(i_haslam,1:n_nu-1),alms(1:n_nu-1,1,ll,mm))
          alms(n_nu,1,ll,mm)=-alms(n_nu,1,ll,mm)/eigenvectors(i_haslam,n_nu)
       END DO
    END DO
  END SUBROUTINE constrain_alms
  
  SUBROUTINE get_galaxy_info(i_pix,t_haslam,specin,amplitude)
    implicit none
    INTEGER(8), intent(in) :: i_pix
    REAL(8), intent(out) :: t_haslam, specin, amplitude

    t_haslam=galaxy_template(i_pix,1)
    specin=specin_template(i_pix,1)
    amplitude=1.0!galaxy_amplitude(i_pix,1)
  END SUBROUTINE get_galaxy_info

  SUBROUTINE end_galaxy_template()
    implicit none
    
    DEALLOCATE(galaxy_template,specin_template,galaxy_amplitude,alm_haslam)
  END SUBROUTINE end_galaxy_template

END MODULE galaxy
