!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                                   !!
!!   Copyright 2012 David Alonso                                     !!
!!                                                                   !!
!!                                                                   !!
!! This file is part of CRIME.                                       !!
!!                                                                   !!
!! CRIME is free software: you can redistribute it and!or modify it  !!
!! under the terms of the GNU General Public License as published by !!
!! the Free Software Foundation, either version 3 of the License, or !!
!! (at your option) any later version.                               !!
!!                                                                   !!
!! CRIME is distributed in the hope that it will be useful, but      !!
!! WITHOUT ANY WARRANTY; without even the implied warranty of        !!
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU !!
!! General Public License for more details.                          !!
!!                                                                   !!
!! You should have received a copy of the GNU General Public License !!
!! along with CRIME.  If not, see <http://www.gnu.org/licenses/>.    !!
!!                                                                   !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE polarization
  USE omp_lib
  USE params_type
  USE math
  USE galaxy, only : NSIDE_TEMPLATE, SPECIN_OFFSET, NU_HASLAM
  USE rngmod
  USE alm_tools
  USE pix_tools
  USE fitstools
  USE udgrade_nr

  implicit none

  INTEGER(4), PARAMETER :: NSIDE_POL=128
  INTEGER(4), PARAMETER :: NPIX_POL=196608 !12*NSIDE*NSIDE
  INTEGER(4), PARAMETER :: NL_POL=383 !3*NSIDE-1
  INTEGER(4), PARAMETER :: N_X=24000
  REAL(8), PARAMETER :: RANGE_X=8.0 !Range in X (in m^2)
  REAL(8), PARAMETER :: X_MIN=-3.0 !Minimum X (in m^2)
  REAL(8), PARAMETER :: CLIGHT=299.792458 !c/(1 MHz) in m
  REAL(8), PARAMETER :: NU_HI=23000.0 !23 GHz

  !These need to be initialized
  REAL(8), PARAMETER :: P_HILAT=0.24 !High latitude depolarization
  REAL(8), PARAMETER :: HILAT_ANGLE=20.0 !Angle defining the NGC
  REAL(8) :: xi_pol !Correlation length in Faraday space
  REAL(8) :: t2mean_hilat_hifreq !mean temperature of the high-frequency map
  REAL(FPP), allocatable, dimension(:,:) :: map_sigma
  REAL(FPP), allocatable, dimension(:,:) :: map_specin !NEW
  REAL(FPP), dimension(0:NL_POL) :: sqrt_cls_pol
  
CONTAINS

  SUBROUTINE end_polarization()
    implicit none

    DEALLOCATE(map_sigma,map_specin)
  END SUBROUTINE end_polarization

  SUBROUTINE init_polarization(pars)
    implicit none
    type(Params4G), intent(in) :: pars
    REAL(FPP), allocatable, dimension(:,:) :: map_T_dum,map_specin_dum
    INTEGER(4), allocatable, dimension(:) :: listpix
    INTEGER(4) :: ll,ii,npix_approx,npix_hilat
    REAL(8) :: tmp
    INTEGER(4) :: stat_alloc

    !Read F-space correlation length
    xi_pol=pars%xi_polarization

    !Read sigma_psi map
    WRITE(*,*) "init_polarization: Reading Faraday widths from file: ",&
         TRIM(pars%fname_faraday)
    ALLOCATE(map_sigma(0:NPIX_POL-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_polarization","map_sigma")
    CALL input_map(pars%fname_faraday,map_sigma,NPIX_POL,1)

    !Read spectral indices
    WRITE(*,*) "init_polarization: Reading spectral indices from file: ",&
         TRIM(pars%fname_specin)
    ALLOCATE(map_specin_dum(0:nside2npix(NSIDE_TEMPLATE)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_polarization","map_specin_dum")
    CALL input_map(pars%fname_specin,map_specin_dum,nside2npix(NSIDE_TEMPLATE),1)
    map_specin_dum(:,1)=map_specin_dum(:,1)-2+SPECIN_OFFSET

    !Read Haslam
    WRITE(*,*) "init_polarization: Reading Haslam map: ",&
         TRIM(pars%fname_haslam)
    ALLOCATE(map_T_dum(0:nside2npix(NSIDE_TEMPLATE)-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_polarization","map_T_dum")
    CALL input_map(pars%fname_haslam,map_T_dum,nside2npix(NSIDE_TEMPLATE),1)
    !These transformations may change for the real map
    CALL convert_inplace(nest2ring,map_T_dum(:,1))
    map_T_dum(:,1)=map_T_dum(:,1)*1000; !Transform from K to mK

    !Calculate mean intensity on high latitudes and frequencies
    npix_approx=INT(nside2npix(NSIDE_TEMPLATE)*0.5*(1-cos(DTOR*HILAT_ANGLE))*1.2) !20% extra just in case
    ALLOCATE(listpix(0:npix_approx),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_polarization","listpix")
    CALL query_strip(NSIDE_TEMPLATE,0.0_8,DTOR*HILAT_ANGLE,listpix,npix_hilat)
    t2mean_hilat_hifreq=0.0
    DO ii=0,npix_hilat-1
       tmp=map_T_dum(listpix(ii),1)*(NU_HI/NU_HASLAM)**map_specin_dum(listpix(ii),1)
       t2mean_hilat_hifreq=t2mean_hilat_hifreq+tmp*tmp
    END DO
    t2mean_hilat_hifreq=t2mean_hilat_hifreq/npix_hilat
    WRITE(*,*) "init_polarization: Mean T at 23 GHz at high latitudes (mK): ",&
         sqrt(t2mean_hilat_hifreq)
    DEALLOCATE(listpix,map_T_dum)

    !Get spectral index map with NSIDE_POL resolution
    ALLOCATE(map_specin(0:NPIX_POL-1,1:1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"init_polarization","map_specin")
    CALL udgrade_ring(map_specin_dum,NSIDE_TEMPLATE,map_specin,NSIDE_POL)
    DEALLOCATE(map_specin_dum)

    !Calculate polarized synchrotron Cl's
    DO ll=0,pars%lmin
       sqrt_cls_pol=(REAL(pars%lref)/pars%lmin)**pars%beta_polarization
       sqrt_cls_pol(ll)=sqrt(sqrt_cls_pol(ll))
    END DO
    DO ll=pars%lmin+1,NL_POL
       sqrt_cls_pol(ll)=(REAL(pars%lref)/ll)**pars%beta_polarization
       sqrt_cls_pol(ll)=sqrt(sqrt_cls_pol(ll))
    END DO
    
  END SUBROUTINE init_polarization

  SUBROUTINE get_uncorrelated_map_single(rng,map_r,map_i)
    implicit none
    type(planck_rng), intent(inout) :: rng
    REAL(FPP), dimension(0:NPIX_POL-1), intent(out) :: map_r
    REAL(FPP), dimension(0:NPIX_POL-1), intent(out) :: map_i
    COMPLEX(FPP), dimension(1:1,0:NL_POL,0:NL_POL) :: alm_aux_r
    COMPLEX(FPP), dimension(1:1,0:NL_POL,0:NL_POL) :: alm_aux_i
    INTEGER(8) :: ll, mm
    REAL(FPP) :: rms
    COMPLEX(8) :: zp,zm

    DO ll=0,NL_POL
       rms=SQRT1HALF*sqrt_cls_pol(ll)
       
       !DO m=0
       zp=rms*CMPLX(rand_gauss(rng),rand_gauss(rng))
       alm_aux_r(1,ll,0)=REAL(zp)
       alm_aux_i(1,ll,0)=AIMAG(zp)
       
       DO mm=1,ll
          zp=rms*CMPLX(rand_gauss(rng),rand_gauss(rng))
          zm=rms*CMPLX(rand_gauss(rng),rand_gauss(rng))
          alm_aux_r(1,ll,mm)=0.5*(zp+zm)
          alm_aux_i(1,ll,mm)=-0.5*ICOM*(zp-zm)
       END DO
    END DO

    CALL alm2map(NSIDE_POL,NL_POL,NL_POL,alm_aux_r,map_r)
    CALL alm2map(NSIDE_POL,NL_POL,NL_POL,alm_aux_i,map_i)
  END SUBROUTINE get_uncorrelated_map_single

  SUBROUTINE get_uncorrelated_maps(seed,map_r,map_i)
    implicit none
    INTEGER(4), intent(in) :: seed
    REAL(FPP), dimension(1:N_X,0:NPIX_POL-1), intent(out) :: map_r
    REAL(FPP), dimension(1:N_X,0:NPIX_POL-1), intent(out) :: map_i
    INTEGER(4) :: ix
    type(planck_rng) :: rng
    INTEGER(4) :: seed_thr,pid

!$omp parallel default(none), &
!$omp private(ix,pid,seed_thr,rng), &
!$omp shared(seed,map_r,map_i,xi_pol)
    pid=omp_get_thread_num()
    seed_thr=seed+pid
    CALL rand_init(rng,seed_thr)

!$omp do
    DO ix=1,N_X
       CALL get_uncorrelated_map_single(rng,map_r(ix,:),map_i(ix,:))
    END DO
!$omp end do
!$omp end parallel
  END SUBROUTINE get_uncorrelated_maps

  SUBROUTINE integrate_LOS_maps(nu,maps_in_r,maps_in_i,map_out_r,map_out_i)
    implicit none
    REAL(8), intent(in) :: nu
    REAL(FPP), dimension(1:N_X,0:NPIX_POL-1), intent(in) :: maps_in_r
    REAL(FPP), dimension(1:N_X,0:NPIX_POL-1), intent(in) :: maps_in_i
    REAL(FPP), dimension(0:NPIX_POL-1), intent(out) :: map_out_r
    REAL(FPP), dimension(0:NPIX_POL-1), intent(out) :: map_out_i
    REAL(8) :: dx,x_lambda,sigma,prefac,kernel,x
    INTEGER(4) :: ipix,ix
    
!$omp parallel default(none), &
!$omp private(ipix,sigma,ix,dx,x_lambda,kernel,prefac,x), &
!$omp shared(map_sigma,map_out_r,map_out_i,maps_in_r,maps_in_i,nu,xi_pol)
    x_lambda=2*(CLIGHT/nu)**2
    dx=RANGE_X/N_X

!$omp do
    DO ipix=0,NPIX_POL-1
       sigma=map_sigma(ipix,1)
       prefac=sqrt(sigma)*dx
       map_out_r(ipix)=0;
       map_out_i(ipix)=0;
       DO ix=1,N_X
          x=X_MIN+(ix-0.5)*dx
          kernel=prefac*exp(-0.5*((x-x_lambda)*sigma)**2-0.25*(x*xi_pol)**2)
          map_out_r(ipix)=map_out_r(ipix)+kernel*maps_in_r(ix,ipix)
          map_out_i(ipix)=map_out_i(ipix)+kernel*maps_in_i(ix,ipix)
       END DO
    END DO
!$omp end do
!$omp end parallel
  END SUBROUTINE integrate_LOS_maps

  SUBROUTINE do_polarization(pars,map_nuspace_r,map_nuspace_i)
    implicit none
    type(Params4G), intent(in) :: pars
    REAL(FPP), dimension(0:,:), intent(out) :: map_nuspace_r
    REAL(FPP), dimension(0:,:), intent(out) :: map_nuspace_i
    REAL(FPP), allocatable, dimension(:,:) :: map_xspace_r
    REAL(FPP), allocatable, dimension(:,:) :: map_xspace_i
    REAL(FPP), allocatable, dimension(:) :: map_aux_r
    REAL(FPP), allocatable, dimension(:) :: map_aux_i
    INTEGER(4) :: ii,ipix,stat_alloc
    REAL(8) :: normalization,q,u,p2,powerlaw,nufrac
    INTEGER(4) :: npix_approx, npix_hilat
    INTEGER(4), allocatable, dimension(:) :: listpix

    REAL(8) :: tbeg,tend

    CALL init_polarization(pars)

    ALLOCATE(map_xspace_r(1:N_X,0:NPIX_POL-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"do_polarization","map_xspace_r")
    ALLOCATE(map_xspace_i(1:N_X,0:NPIX_POL-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"do_polarization","map_xspace_i")

    tbeg=omp_get_wtime()
    WRITE(*,*) "do_polarization: calculating uncorrelated maps"
    WRITE(*,*) "nthreads = ", omp_get_max_threads()
    CALL get_uncorrelated_maps(pars%seed,map_xspace_r,map_xspace_i)
    tend=omp_get_wtime()
    WRITE(*,*) "do_polarization: time ellapsed (s): ", tend-tbeg

    tbeg=omp_get_wtime()
    WRITE(*,*) "do_polarization: integranting LOSs"
    DO ii=1,pars%n_nu
       CALL integrate_LOS_maps(pars%nutable(ii,1),map_xspace_r,map_xspace_i,&
            map_nuspace_r(:,ii),map_nuspace_i(:,ii))
    END DO
    tend=omp_get_wtime()
    WRITE(*,*) "do_polarization: time ellapsed (s): ", tend-tbeg

    WRITE(*,*) "do_polarization: normalizing maps"
    ALLOCATE(map_aux_r(0:NPIX_POL-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"do_polarization","map_aux_r")
    ALLOCATE(map_aux_i(0:NPIX_POL-1),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"do_polarization","map_aux_r")
    WRITE(*,*) "do_polarization: calculating high-frequency map"
    CALL integrate_LOS_maps(NU_HI,map_xspace_r,map_xspace_i,&
         map_aux_r,map_aux_i)
    DEALLOCATE(map_xspace_r,map_xspace_i)

    WRITE(*,*) "do_polarization: fixing high latitude polarized fraction at ",P_HILAT
    npix_approx=INT(NPIX_POL*0.5*(1-cos(DTOR*HILAT_ANGLE))*1.2) !20% extra just in case
    ALLOCATE(listpix(0:npix_approx),STAT=stat_alloc)
    CALL assert_alloc(stat_alloc,"do_polarization","listpix")
    CALL query_strip(NSIDE_POL,0.0_8,DTOR*HILAT_ANGLE,listpix,npix_hilat)
    normalization=0.0
    DO ii=0,npix_hilat-1
       q=map_aux_r(listpix(ii))
       u=map_aux_i(listpix(ii))
       p2=q*q+u*u
       normalization=normalization+p2
    END DO
    normalization=normalization/npix_hilat
    normalization=sqrt(P_HILAT**2/(1-P_HILAT**2)*(t2mean_hilat_hifreq/normalization))
    DEALLOCATE(listpix,map_aux_r,map_aux_i)

    DO ii=1,pars%n_nu
       nufrac=pars%nutable(ii,1)/NU_HI
       DO ipix=0,NPIX_POL-1
          powerlaw=nufrac**map_specin(ipix,1)
          map_nuspace_r(ipix,ii)=powerlaw*normalization*map_nuspace_r(ipix,ii)
          map_nuspace_i(ipix,ii)=powerlaw*normalization*map_nuspace_i(ipix,ii)
       END DO
    END DO

  END SUBROUTINE do_polarization

END MODULE polarization
