///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of CosmoMAD.                                    //
//                                                                   //
// CosmoMAD is free software: you can redistribute it and/or modify  //
// it under the terms of the GNU General Public License as published //
// by the Free Software Foundation, either version 3 of the License, //
// or (at your option) any later version.                            //
//                                                                   //
// CosmoMAD is distributed in the hope that it will be useful, but   //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with CosmoMAD.  If not, see <http://www.gnu.org/licenses/>. //
//                                                                   //
///////////////////////////////////////////////////////////////////////

/**
 * @file cosmo_mad.h
 * @author David Alonso
 * @date 13 Oct 2013
 * @brief Header file for the CosmoMAD library
 *
 * This header file contains the definitions of all
 * the available CosmoMAD functions and macros.
 */

#ifndef _COSMO_MAD_H
#define _COSMO_MAD_H

#include <gsl/gsl_integration.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_sf_gamma.h>

#define CSM_FOURPITHIRD 4.18879020478639 //4*pi/3
#define CSM_LOGTEN 2.302585092994 //ln(10)
#define CSM_RTOD 57.29577951308232 //180/pi
#define CSM_DTOR 0.017453292519943295 //pi/180
#define CSM_HGYR 9.777751486751187 //H0^-1 in Gyrs/h
#define CSM_HMPC 2997.92458 //H0^-1 in Mpc/h

#define CSM_MIN(a, b)  (((a) < (b)) ? (a) : (b))
#define CSM_MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define CSM_ABS(a)   (((a) < 0) ? -(a) : (a))
#define CSM_CLAMP(x, low, high)  (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))
// min(max(x,low),high)

typedef struct {
  double rsound;
  double zeq;
  double keq;
  double zdrag;
  double kSilk;
  double rsound_approx;
  double th2p7;
  double alphac;
  double alphab;
  double betac;
  double betab;
  double bnode;
} Csm_EH_params;

typedef struct {
  //Tuneable parameters
  double OM;
  double OB;
  double OL;
  double h;
  double w0;
  double wa;
  double TCMB;

  // Derived parameters
  double OK;
  int ksign;
  int normalDE;
  int constantw;
  double bang_time;
  double phorizon;
  double a_equality;
  double growth0;
  Csm_EH_params *eh;

  // a(t) spline
  int at_spline_set;
  gsl_interp_accel *intacc_at;
  gsl_spline *spline_at;
} Csm_bg_params;

/**
 * @brief Basic parameter structure in CosmoMAD.
 *
 * A Csm_params structure contains all the information
 * loaded by the user regarding a given cosmological model.
 */
typedef struct {
  /* Background parameters */
  int bg_params_set;
  Csm_bg_params *bg;
} Csm_params;


/****************************/
/*   Background cosmology   */
/****************************/
/**
 * @brief Unset GSL error handler
 *
 * Disables the default GSL error handler. The user will then be
 * notified if any errors or warnings are found regarding the
 * GSL functions (e.g.: some integral was not able to reach the
 * required precision). These warnings will often be unimportant,
 * and the code will not exit.
 */
void csm_unset_gsl_eh(void);

/**
 * @brief Sets verbosity level
 *
 * Determines the amount of information to be output. Only two
 * values for @p verb are supported: 0 (nothing) and 1 (everything).
 * The default verbosity level is 1 (all messages are shown).
 */
void csm_set_verbosity(int verb);

/**
 * @brief Csm_params destructor.
 *
 * Frees up all memory associated with @p par.
 */
void csm_params_free(Csm_params *par);

/**
 * @brief Csm_params creator.
 *
 * Returns an initialized Csm_params structure, without any
 * associated cosmological information (see #csm_background_set).
 */
Csm_params *csm_params_new(void);

/**
 * @brief Hubble parameter
 *
 * Returns the Hubble parameter \f$H(a)\f$ at 
 * \f$a = {\tt aa}\f$ in \f$h/\f$Mpc.
 */
double csm_hubble(Csm_params *par,double aa);

/**
 * @brief Cosmic time
 *
 * Returns the cosmic time \f$t(a)\f$ at \f$a={\tt aa}\f$
 * by calculating the integral
 * \f[
 *    t(a) = \int_0^a \frac{da'}{a'\,H(a')}.
 * \f]
 */
double csm_cosmic_time(Csm_params *par,double aa);

/**
 * @brief Particle horizon
 *
 * Returns the particle horizon \f$\chi_p(a)\f$ at
 * \f$a={\tt aa}\f$ by calculating the integral 
 * \f[
 *    \chi_p(a) \int_0^a\frac{da'}{a'^2\,H(a')}.
 * \f]
 */
double csm_particle_horizon(Csm_params *par,double aa);

/**
 * @brief Radial comoving distance
 *
 * Returns the radial comoving distance \f$\chi(a)\f$ 
 * at \f$a={\tt aa}\f$ as
 * \f$\chi(a)=\chi_p(1)-\chi_p(a)\f$, where \f$\chi_p\f$ is
 * the particle horizon calculated with
 * #csm_particle_horizon.
 */
double csm_radial_comoving_distance(Csm_params *par,double aa);

/**
 * @brief Curvature comoving distance
 *
 * Returns the curvature comoving distance \f$r(a)\f$ 
 * at \f$a={\tt aa}\f$ as
 * \f[
 *   r(a) =  \frac{1}{H_0\sqrt{|\Omega_k|}}{\rm sinn}\left(H_0\sqrt{|\Omega_k|}\chi(a)\right),
 * \f]
 * where \f$\chi(a)\f$ is the radial comoving distance calculated
 * with #csm_radial_comoving_distance.
 */
double csm_curvature_comoving_distance(Csm_params *par,double aa);

/**
 * @brief Angular diameter distance
 *
 * Returns the angular diameter distance \f$d_A(a)\f$ 
 * at \f$a={\tt aa}\f$ as \f$d_A(a) = a\,r(a)\f$, where
 * \f$r(a)\f$ is the curvature comoving distance calculated
 * with #csm_curvature_comoving_distance.
 */
double csm_angular_diameter_distance(Csm_params *par,double aa);

/**
 * @brief Luminosity distance
 *
 * Returns the luminosity distance \f$d_L(a)\f$ 
 * at \f$a={\tt aa}\f$ as \f$d_L(a) = r(a)/a\f$, where
 * \f$r(a)\f$ is the curvature comoving distance calculated
 * with #csm_curvature_comoving_distance.
 */
double csm_luminosity_distance(Csm_params *par,double aa);

/**
 * @brief Growth factor
 *
 * Returns the linear growth factor \f$D(a)\f$ at
 * \f$a={\tt aa}\f$ (normalized to \f$D(a\ll1)\simeq a\f$)
 * by calculating the integral
 * \f[
 *   D(a) = \frac{5}{2}\Omega_M\,H_0^2\,H(a)
 *          \int_0^a\frac{da'}{[a'\,H(a')]^3}.
 * \f]
 */
double csm_growth_factor(Csm_params *par,double aa);

/**
 * @brief Growth rate
 *
 * Returns the linear growth rate \f$f(a)\equiv 
 * d\,\log D(a)/d\,\log a\f$ at \f$a={\tt aa}\f$.
 */
double csm_f_growth(Csm_params *par,double aa);

/**
 * @brief Scale factor
 *
 * Returns the value of the scale factor \f$a(t)\f$
 * at cosmic time \f$t={\tt t}\f$. The first time this
 * function is called, the relation \f$t(a)\f$ is calculated
 * for several values of \f$a\in[0,1]\f$ using #csm_cosmic_time,
 * and a spline is used to invert this relation in all subsequent
 * calls.
 */
double csm_scale_factor(Csm_params *par,double t);

/**
 * @brief Angular BAO scale
 * 
 * Returns the angular scale of the BAO (in deg) at
 * \f$a = {\tt aa}\f$ as 
 * \f[
 *   \theta_{\rm BAO}(a) = \frac{r_s}{r(a)},
 * \f]
 * where \f$r_s\f$ is the sound horizon scale estimated
 * through the Eiseinstein \& Hu fitting formula and
 * \f$r(a)\f$ is the curvature comoving distance calculated
 * with #csm_curvature_comoving_distance.
 */
double csm_theta_BAO(Csm_params *par,double aa);

/**
 * @brief Radial BAO scale
 * 
 * Returns the radial scale of the BAO as a redshift separation
 * \f$a = {\tt aa}\f$ as \f$\Delta z_{\rm BAO}(a) = H(a)\,r_s\f$,
 * where \f$r_s\f$ is the sound horizon scale estimated
 * through the Eiseinstein \& Hu fitting formula and
 * \f$H(a)\f$ is the Hubble paramteter calculated with #csm_hubble.
 */
double csm_Dz_BAO(Csm_params *par,double aa);

/**
 * @brief Set background cosmology
 *
 * Sets background cosmology for @p par: \f$\Omega_M={\tt OM}\f$,
 * \f$\Omega_{\Lambda}={\tt OL}\f$, \f$\Omega_b={\tt OB}\f$,
 * \f$w_0={\tt w0}\f$, \f$w_a={\tt wa}\f$, \f$h={\tt hh}\f$
 * and \f$T_{\rm CMB}={\tt T\_CMB}\f$. This function must be
 * called before calculating any \f$a\f$-dependent quantity.
 */
void csm_background_set(Csm_params *par,
			double OmegaM,double OmegaL,double OmegaB,
			double ww,double wwa,double hh,double T_CMB);

#endif //_COSMO_MAD_
