///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of GetHI.                                       //
//                                                                   //
// GetHI is free software: you can redistribute it and/or modify it  //
// under the terms of the GNU General Public License as published by //
// the Free Software Foundation, either version 3 of the License, or //
// (at your option) any later version.                               //
//                                                                   //
// GetHI is distributed in the hope that it will be useful, but      //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with GetHI.  If not, see <http://www.gnu.org/licenses/>.    //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
#include <gsl/gsl_rng.h>

#ifdef _HAVE_OMP
#include <omp.h>
#endif //_HAVE_OMP

#include "fftw3.h"
#include "common.h"

#ifdef _SPREC
typedef fftwf_complex dftw_complex;
#else //_SPREC
typedef fftw_complex dftw_complex;
#endif //_SPREC

static void init_fftw(void)
{
  //////
  // Initializes threads if using OpenMP
#ifdef _HAVE_OMP
#ifdef _SPREC
  int stat=fftwf_init_threads();
#else //_SPREC
  int stat=fftw_init_threads();
#endif //_SPREC
  if(!stat) {
    fprintf(stderr,"GetHI: Couldn't initialize FFTW threads \n");
    exit(1);
  }

#ifdef _SPREC
  fftwf_plan_with_nthreads(omp_get_max_threads());
#else //_SPREC
  fftw_plan_with_nthreads(omp_get_max_threads());
#endif //_SPREC

#endif //_HAVE_OMP
}

static void end_fftw(void)
{
  //////
  // Frees up FFTW memory
#ifdef _HAVE_OMP
#ifdef _SPREC
  fftwf_cleanup_threads();
#else //_SPREC
  fftw_cleanup_threads();
#endif //_SPREC
#else //_HAVE_OMP
#ifdef _SPREC
  fftwf_cleanup();
#else //_SPREC
  fftw_cleanup();
#endif //_SPREC
#endif //_HAVE_OMP
}

static dftw_complex *create_density_fourier(unsigned int *seed)
{
  //////
  // Generates a random realization of the delta_k
  // from the linear P_k.
  dftw_complex *density;
  lint n_grid_tot=n_grid*n_grid;
  n_grid_tot*=(n_grid/2+1);

#ifdef _SPREC
  density=fftwf_alloc_complex(n_grid_tot);
#else //_SPREC
  density=fftw_alloc_complex(n_grid_tot);
#endif //_SPREC
  if(density==NULL) error_mem_out();

#pragma omp parallel default(none)		\
  shared(n_grid,seed,density,l_box)
  {
    int ii;
    double dk=2*M_PI/l_box;
    double idk3=1./(dk*dk*dk);
    unsigned int seed_thr=*seed+omp_get_thread_num();
    gsl_rng *rng_thr=init_rng(seed_thr);
    
#pragma omp for schedule(dynamic,DYNAMIC_SIZE)
    for(ii=0;ii<n_grid;ii++) {
      int jj;
      double kz;
      if(2*ii<=n_grid)
	kz=ii*dk;
      else
	kz=-(n_grid-ii)*dk;
      for(jj=0;jj<n_grid;jj++) {
	int kk;
	double ky;
	if(2*jj<=n_grid)
	  ky=jj*dk;
	else
	  ky=-(n_grid-jj)*dk;
	for(kk=0;kk<=n_grid/2;kk++) {
	  double kx;
	  double k_mod2;
	  lint index=kk+(n_grid/2+1)*((lint)(jj+n_grid*ii)); //Grid index for +k
	  double sigma2;
	  double delta_mod,delta_phase;
	  if(2*kk<=n_grid)
	    kx=kk*dk;
	  else
	    kx=-(n_grid-kk)*dk; //This should never happen
	  
	  k_mod2=kx*kx+ky*ky+kz*kz;
	  
	  if(k_mod2>0) {
	    double lgk=0.5*log10(k_mod2);
	    sigma2=pk_linear0(lgk)*idk3;
	  }
	  else
	    sigma2=0;
	  
	  rng_delta_gauss(&delta_mod,&delta_phase,rng_thr,sigma2);
	  density[index]=delta_mod*cexp(I*delta_phase);
	}
      }
    }
    end_rng(rng_thr);
  }

  return density;
}

static dftw_complex *velocity_potential_from_delta(dftw_complex *delta)
{
  dftw_complex *vel_phi;
  lint n_grid_tot=n_grid*n_grid;
  n_grid_tot*=n_grid;

#ifdef _SPREC
  vel_phi=fftwf_alloc_complex(n_grid_tot);
#else //_SPREC
  vel_phi=fftw_alloc_complex(n_grid_tot);
#endif //_SPREC
  if(vel_phi==NULL) error_mem_out();

#pragma omp parallel default(none)			\
  shared(vel_phi,delta,l_box,n_grid,fgrowth_0,hubble_0)
  {
    int ii;
    double dk=2*M_PI/l_box;
    double factor=fgrowth_0*hubble_0;

#pragma omp for schedule(dynamic,DYNAMIC_SIZE)
    for(ii=0;ii<n_grid;ii++) {
      int jj;
      double kz;
      if(2*ii<=n_grid)
	kz=ii*dk;
      else
	kz=-(n_grid-ii)*dk;
      for(jj=0;jj<n_grid;jj++) {
	int kk;
	double ky;
	if(2*jj<=n_grid)
	  ky=jj*dk;
	else
	  ky=-(n_grid-jj)*dk;
	for(kk=0;kk<=n_grid/2;kk++) {
	  double kx;
	  double k_mod2;
	  lint index=kk+(n_grid/2+1)*
	    ((lint)(jj+n_grid*ii)); //Grid index for +k
	  if(2*kk<=n_grid)
	    kx=kk*dk;
	  else
	    kx=-(n_grid-kk)*dk; //This should never happen
	  
	  k_mod2=kx*kx+ky*ky+kz*kz;
	  
	  if(k_mod2==0) vel_phi[index]=0;
	  else
	    vel_phi[index]=delta[index]*factor/k_mod2;
	}
      }
    } // end omp for
  } // end omp parallel

  return vel_phi;
}

static flouble *radial_velocity_from_potential(flouble *vel_phi)
{
  flouble *vel_rad;
  lint n_grid_tot=n_grid*n_grid;
  n_grid_tot*=n_grid;
  
  vel_rad=(flouble *)malloc(n_grid_tot*sizeof(flouble));
  if(vel_rad==NULL) error_mem_out();

#pragma omp parallel default(none)		\
  shared(n_grid,l_box,vel_phi,vel_rad,pos_obs)
  {
    double dx=l_box/n_grid;
    double idx=1./dx;
    int iz;
    
#pragma omp for schedule(dynamic,DYNAMIC_SIZE)
    for(iz=0;iz<n_grid;iz++) {
      int iy;
      lint iz_hi=iz+1;
      lint iz_lo=iz-1;
      lint iz_0=iz;
      double z=dx*(iz+0.5)-pos_obs[2];
      if(iz==0) iz_lo=n_grid-1;
      if(iz==n_grid-1) iz_hi=0;
      iz_hi*=n_grid*n_grid;
      iz_lo*=n_grid*n_grid;
      iz_0*=n_grid*n_grid;
      for(iy=0;iy<n_grid;iy++) {
	int ix;
	lint iy_hi=iy+1;
	lint iy_lo=iy-1;
	lint iy_0=iy;
	double y=dx*(iy+0.5)-pos_obs[1];
	if(iy==0) iy_lo=n_grid-1;
	if(iy==n_grid-1) iy_hi=0;
	iy_hi*=n_grid;
	iy_lo*=n_grid;
	iy_0*=n_grid;
	for(ix=0;ix<n_grid;ix++) {
	  double vel[3];
	  double ur[3];
	  lint ix_hi=ix+1;
	  lint ix_lo=ix-1;
	  lint ix_0=ix;
	  double x=dx*(ix+0.5)-pos_obs[0];
	  double irr=1./sqrt(x*x+y*y+z*z);
	  if(ix==0) ix_lo=n_grid-1;
	  if(ix==n_grid-1) ix_hi=0;
	  
	  ur[0]=x*irr;
	  ur[1]=y*irr;
	  ur[2]=z*irr;
	  
	  vel[0]=0.5*idx*(vel_phi[ix_hi+iy_0+iz_0]-vel_phi[ix_lo+iy_0+iz_0]);
	  vel[1]=0.5*idx*(vel_phi[ix_0+iy_hi+iz_0]-vel_phi[ix_0+iy_lo+iz_0]);
	  vel[2]=0.5*idx*(vel_phi[ix_0+iy_0+iz_hi]-vel_phi[ix_0+iy_0+iz_lo]);

	  vel_rad[ix_0+iy_0+iz_0]=vel[0]*ur[0]+vel[1]*ur[1]+vel[2]*ur[2];
	}
      }
    } // end omp for
  } // end omp parallel
  
  return vel_rad;
}

void create_d_and_vr_fields(unsigned int *seed,flouble **delta,
			    flouble **radial_velocity)
{
  //////
  // Creates a realization of the gaussian density
  // contrast field from the linear P_k
#ifdef _SPREC
  fftwf_plan plan_ft;
#else //_SPREC
  fftw_plan plan_ft;
#endif //_SPREC

  dftw_complex *density;
  flouble *density_result;

  dftw_complex *vel_phi=NULL;
  flouble *vel_phi_result=NULL;
  flouble *vel_rad=NULL;
  
  lint n_grid_tot=n_grid*n_grid;
#ifdef _FFT_IN_PLACE
  n_grid_tot*=2*(n_grid/2+1);
#else //_FFT_IN_PLACE
  n_grid_tot*=n_grid;
#endif //_FFT_IN_PLACE
  printf("*** Creating Gaussian density field \n");

  init_fftw();

  printf("Creating Fourier-space density \n");
  timer(0);
  density=create_density_fourier(seed);
#ifdef _FFT_IN_PLACE
  density_result=(flouble *)density;
#else //_FFT_IN_PLACE
#ifdef _SPREC
  density_result=fftwf_alloc_real(n_grid_tot);
#else //_SPREC
  density_result=fftw_alloc_real(n_grid_tot);
#endif //_SPREC
  if(density_result==NULL) error_mem_out();
#endif //_FFT_IN_PLACE
  timer(2);

  printf("Creating Fourier-space velocity potential \n");
  timer(0);
  vel_phi=velocity_potential_from_delta(density);
#ifdef _FFT_IN_PLACE
  vel_phi_result=(flouble *)vel_phi;
#else //_FFT_IN_PLACE
#ifdef _SPREC
  vel_phi_result=fftwf_alloc_real(n_grid_tot);
#else //_SPREC
  vel_phi_result=fftw_alloc_real(n_grid_tot);
#endif //_SPREC
  if(vel_phi_result==NULL) error_mem_out();
#endif //_FFT_IN_PLACE
  timer(2);

  printf("Transforming density\n");
  timer(0);
#ifdef _SPREC
  plan_ft=fftwf_plan_dft_c2r_3d(n_grid,n_grid,n_grid,density,
			       density_result,FFTW_ESTIMATE);
  fftwf_execute(plan_ft);
  fftwf_destroy_plan(plan_ft);
#ifndef _FFT_IN_PLACE
  fftwf_free(density);
#endif //_FFT_IN_PLACE
#else //_SPREC
  plan_ft=fftw_plan_dft_c2r_3d(n_grid,n_grid,n_grid,density,
			       density_result,FFTW_ESTIMATE);
  fftw_execute(plan_ft);
  fftw_destroy_plan(plan_ft);
#ifndef _FFT_IN_PLACE
  fftw_free(density);
#endif //_FFT_IN_PLACE
#endif //_SPREC

  timer(2);

  printf("Normalizing density \n");
  timer(0);
#pragma omp parallel default(none)				\
  shared(density_result,n_grid,n_grid_tot,l_box)
  {
    lint ii;
    double norm=pow(sqrt(2*M_PI)/l_box,3);
    
#pragma omp for
    for(ii=0;ii<n_grid_tot;ii++)
      density_result[ii]*=norm;
  }//end omp parallel
  timer(2);

  printf("Transforming velocity potential\n");
  timer(0);
#ifdef _SPREC
  plan_ft=fftwf_plan_dft_c2r_3d(n_grid,n_grid,n_grid,vel_phi,
				vel_phi_result,FFTW_ESTIMATE);
  fftwf_execute(plan_ft);
  fftwf_destroy_plan(plan_ft);
#ifndef _FFT_IN_PLACE
  fftwf_free(vel_phi);
#endif //_FFT_IN_PLACE
#else //_SPREC
  plan_ft=fftw_plan_dft_c2r_3d(n_grid,n_grid,n_grid,vel_phi,
			       vel_phi_result,FFTW_ESTIMATE);
  fftw_execute(plan_ft);
  fftw_destroy_plan(plan_ft);
#ifndef _FFT_IN_PLACE
  fftw_free(vel_phi);
#endif //_FFT_IN_PLACE
#endif //_SPREC

  timer(2);
  
  printf("Normalizing velocity potential \n");
  timer(0);
#pragma omp parallel default(none)			\
  shared(vel_phi_result,n_grid,n_grid_tot,l_box)
  {
    lint ii;
    double norm=pow(sqrt(2*M_PI)/l_box,3);
    
#pragma omp for
    for(ii=0;ii<n_grid_tot;ii++)
      vel_phi_result[ii]*=norm;
  }//end omp parallel
  timer(2);
  
#ifdef _FFT_IN_PLACE
  printf("Undoing padding\n");
  timer(0);
  int i;
  for(i=0;i<n_grid;i++) {
    int j;
    for(j=0;j<n_grid;j++) {
      int k;
      for(k=0;k<n_grid;k++) {
	lint index_unpadded=k+n_grid*((lint)(j+n_grid*i));
	lint index_padded=k+2*(n_grid/2+1)*((lint)(j+n_grid*i));
	density_result[index_unpadded]=density_result[index_padded];
	vel_phi_result[index_unpadded]=vel_phi_result[index_padded];
      }
    }
  }
  timer(2);
#endif //_FFT_IN_PLACE

  printf("Calculating radial velocity \n");
  timer(0);
  vel_rad=radial_velocity_from_potential(vel_phi_result);
#ifdef _SPREC
  fftwf_free(vel_phi_result);
#else //_SPRC
  fftw_free(vel_phi_result);
#endif //_SPRC
  timer(2);

  end_fftw();

  *delta=density_result;
  *radial_velocity=vel_rad;

  printf("\n");
}
