///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of GetHI.                                       //
//                                                                   //
// GetHI is free software: you can redistribute it and/or modify it  //
// under the terms of the GNU General Public License as published by //
// the Free Software Foundation, either version 3 of the License, or //
// (at your option) any later version.                               //
//                                                                   //
// GetHI is distributed in the hope that it will be useful, but      //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with GetHI.  If not, see <http://www.gnu.org/licenses/>.    //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_rng.h>

#ifdef _HAVE_OMP
#include <omp.h>
#endif //_HAVE_OMP

#include "common.h"

int *get_point_sources(flouble *delta_grid,unsigned int *seed)
{
  //////
  // Uses the gaussian matter density field to obtain a
  // poisson sampling of point sources (returns an integer array
  // with the number of sources in each cell).
  double sigma2_gauss=0;
  double delta_mean=0;
  lint n_grid_tot=n_grid*((lint)(n_grid*n_grid));

  printf("*** Getting point sources\n");

  timer(0);
#pragma omp parallel default(none)				\
  shared(delta_grid,n_grid,sigma2_gauss,delta_mean,n_grid_tot)
  {
    lint ii;
    double sigma2_thr=0;
    double delta_mean_thr=0;

#pragma omp for nowait
    for(ii=0;ii<n_grid_tot;ii++) {
      sigma2_thr+=delta_grid[ii]*delta_grid[ii];
      delta_mean_thr+=delta_grid[ii];
    } //end pragma omp for

#pragma omp critical
    {
      delta_mean+=delta_mean_thr/n_grid_tot;
      sigma2_gauss+=sigma2_thr/n_grid_tot;
    } //end pragma omp critical
  } //end pragma omp parallel
  
  sigma2_gauss-=delta_mean*delta_mean;
#ifdef _DEBUG
  printf("  Average overdensity = %lE %lE \n",
	 delta_mean,sqrt(sigma2_gauss));
#endif //_DEBUG
  
  int *nsources=(int *)malloc(n_grid_tot*sizeof(double));
  if(nsources==NULL) error_mem_out();
  printf("Poisson-sampling\n");
  lint np_tot=0;
#pragma omp parallel default(none)		\
  shared(delta_grid,seed,n_grid,l_box,pos_obs)	\
  shared(sigma2_gauss,nsources,np_tot)
  {
    int iz;
    double dx=l_box/n_grid;
    double cell_vol=dx*dx*dx;
    lint np_tot_thr=0;
    unsigned int seed_thr=*seed+omp_get_thread_num();
    gsl_rng *rng_thr=init_rng(seed_thr);

#pragma omp for schedule(dynamic,DYNAMIC_SIZE)
    for(iz=0;iz<n_grid;iz++) {
      int iy;
      lint indexz=(lint)(iz*n_grid)*n_grid;
      double z0=(iz+0.5)*dx-pos_obs[2];
      for(iy=0;iy<n_grid;iy++) {
	int ix;
	lint indexy=(lint)(iy*n_grid);
	double y0=(iy+0.5)*dx-pos_obs[1];
	for(ix=0;ix<n_grid;ix++) {
	  lint index=ix+indexy+indexz;
	  double x0=(ix+0.5)*dx-pos_obs[1];
	  double r=sqrt(x0*x0+y0*y0+z0*z0);
	  double redshift=z_of_r(r);
	  double ndens=n_of_z_psources(redshift);
	  int npp;
	  if(ndens>0) {
	    double bias=bias_psources(redshift);
	    double gfb=dgrowth_of_r(r)*bias;
	    double lambda=ndens*cell_vol*
	      exp(gfb*(delta_grid[index]-0.5*gfb*sigma2_gauss));
	    npp=rng_poisson(lambda,rng_thr);
	  }
	  else npp=0;

	  nsources[index]=npp;
	  np_tot_thr+=npp;
	}
      }
    }//end omp for

#pragma omp critical
    {
      np_tot+=np_tot_thr;
    }//end omp critical
    end_rng(rng_thr);
  }//end omp parallel
  timer(2);

#ifdef _LONGIDS
  printf("  There will be %ld particles in total \n",np_tot);
#else //LONGIDS
  printf("  There will be %d particles in total \n",np_tot);
#endif //_LONGIDS

  return nsources;
}

void get_HI(flouble *delta_grid,flouble *v_rad)
{
  //////
  // 1 - Applies lognormal transformation to density grid
  // 2 - Transforms radial velocity into redshift distortion
  // 3 - Substitutes the overdensity grid for the corresponding
  //     HI mass in each cell.

  double sigma2_gauss=0;
  double delta_mean=0;

  printf("*** Gettin' HI\n");

  timer(0);
#pragma omp parallel default(none)				\
  shared(delta_grid,v_rad,n_grid,z_min,z_max)			\
  shared(sigma2_gauss,delta_mean,l_box,pos_obs,OmegaB,hhub)
  {
    lint ii;
    double sigma2_thr=0;
    double delta_mean_thr=0;
    double dx=l_box/n_grid;
    //Using this the mass grid will be in units of M_sun
    //    double mass_prefac=2.77459457E11*OmegaB*dx*dx*dx/hhub;
    //Using this the mass grid will be in units of 2.77459457E11*OmegaB/hhub M_sun
    double mass_prefac=dx*dx*dx;
    lint n_grid_tot=n_grid*n_grid;
    n_grid_tot*=n_grid;
    

#pragma omp for nowait
    for(ii=0;ii<n_grid_tot;ii++) {
      sigma2_thr+=delta_grid[ii]*delta_grid[ii];
      delta_mean_thr+=delta_grid[ii];
    }

#pragma omp critical
    {
      delta_mean+=delta_mean_thr/n_grid_tot;
      sigma2_gauss+=sigma2_thr/n_grid_tot;
    } //end pragma omp critical
#pragma omp barrier //Synchronize before lognormalizing
#pragma omp single
    {
      sigma2_gauss-=delta_mean*delta_mean;
    }
#pragma omp barrier
#ifdef _DEBUG
#pragma omp single
    {
      printf("  Average overdensity = %lE %lE \n",
	     delta_mean,sqrt(sigma2_gauss));
    }
#endif //_DEBUG

    int iz;
#pragma omp for nowait
    for(iz=0;iz<n_grid;iz++) {
      double z=dx*(iz+0.5)-pos_obs[2];
      int iy;
      for(iy=0;iy<n_grid;iy++) {
	double y=dx*(iy+0.5)-pos_obs[1];
	int ix;
	for(ix=0;ix<n_grid;ix++) {
	  lint index=ix+n_grid*((lint)(iy+n_grid*iz));
	  double x=dx*(ix+0.5)-pos_obs[0];
	  double r=sqrt(x*x+y*y+z*z);
	  double redshift=z_of_r(r);
	  double gfd=dgrowth_of_r(r)*bias_HI(redshift);
	  double delta_gauss=delta_grid[index];
	  double dz_rsd=v_rad[index]*vgrowth_of_r(r);
	  double dens_LN=exp(gfd*(delta_gauss-0.5*gfd*sigma2_gauss));
	  double mass_HI=mass_prefac*fraction_HI(redshift)*dens_LN;
	  delta_grid[index]=mass_HI;
	  v_rad[index]=dz_rsd;
	}
      }
    } //end pragma omp for
  } //end pragma omp parallel
  timer(2);
  printf("\n");
}
