///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of GetHI.                                       //
//                                                                   //
// GetHI is free software: you can redistribute it and/or modify it  //
// under the terms of the GNU General Public License as published by //
// the Free Software Foundation, either version 3 of the License, or //
// (at your option) any later version.                               //
//                                                                   //
// GetHI is distributed in the hope that it will be useful, but      //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with GetHI.  If not, see <http://www.gnu.org/licenses/>.    //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_rng.h>

#ifdef _HAVE_OMP
#include <omp.h>
#endif //_HAVE_OMP

#include "chealpix.h"
#include "common.h"

#define N_SUBPART 10

#define LUM2FLUX 8.35774E7 //(10^22 Joules)/(4*pi*(1 Mpc)^2)/[1 microJy]
#define FLUX2TEMP 3.2548291E-2 //c^2*(1 microJy)/(2*kB*(1 MHz)^2)/[1 mK]
#define FLUXMIN 1E-2 //Smin = 10 nJy
#define FLUXCUT 1E7 //Smax = 10 Jy
#define NU0_PSOURCES 1420.

flouble *mk_psources_maps(unsigned int *seed,int *nsources,flouble *dz_grid)
{
  int ii;
  flouble *T_maps;
  long n_pix_total=nside2npix(n_side)*n_nu;
  double *nu_arr=(double *)malloc(n_nu*sizeof(double));
  if(nu_arr==NULL) error_mem_out();
  
  for(ii=0;ii<n_nu;ii++)
    nu_arr[ii]=nu_min+(nu_max-nu_min)*(ii+0.5)/n_nu;

  printf("*** Making source maps\n");
  T_maps=(flouble *)calloc(n_pix_total,sizeof(flouble));
  if(T_maps==NULL) error_mem_out();

  timer(0);

  //dbg
  long np_total=0;
#pragma omp parallel default(none)		\
  shared(nsources,dz_grid,n_grid,T_maps,hhub)	\
  shared(l_box,pos_obs,n_side,seed,nu_arr,n_nu)	\
  shared(np_total)
  {
    int iz;
    double dx=l_box/n_grid;
    unsigned int seed_thr=*seed+omp_get_thread_num();
    gsl_rng *rng_thr=init_rng(seed_thr);
    double s_prefactor=LUM2FLUX*hhub*hhub;
    long n_pix_ang=nside2npix(n_side);
    double dOmega=4*M_PI/n_pix_ang;
    double t_prefactor=FLUX2TEMP/dOmega;
    
    //dbg
    long np_thr=0;

#pragma omp for schedule(dynamic,DYNAMIC_SIZE)
    for(iz=0;iz<n_grid;iz++) {
      int iy;
      lint indexz=iz*((lint)(n_grid*n_grid));
      double z0=dx*iz-pos_obs[2];
      for(iy=0;iy<n_grid;iy++) {
	int ix;
	lint indexy=(lint)(iy*n_grid);
	double y0=dx*iy-pos_obs[1];
	for(ix=0;ix<n_grid;ix++) {
	  lint index=ix+indexy+indexz;
	  double x0=dx*ix-pos_obs[0];
	  int np=nsources[index];
	  if(np>0) {
	    int ip;
	    double dz_rsd=(double)(dz_grid[index]);
	    for(ip=0;ip<np;ip++) {
	      double pos[3];
	      double r,red_cosmo,red_true;
	      double l_nu,s_nu;
	      pos[0]=x0+dx*rng_01(rng_thr);
	      pos[1]=y0+dx*rng_01(rng_thr);
	      pos[2]=z0+dx*rng_01(rng_thr);
	      r=sqrt(pos[0]*pos[0]+pos[1]*pos[1]+pos[2]*pos[2]);
	      red_cosmo=z_of_r(r);
	      red_true=red_cosmo+dz_rsd;
	      //Luminosity in units of 10^22 W/Hz
	      l_nu=draw_luminosity(red_cosmo,rng_thr);
	      //Flux density in microJy
	      s_nu=s_prefactor*l_nu/(r*r*(1+red_true));
	      if(s_nu>=FLUXMIN) {
		int inu;
		double specin;
		long ipix;

		//dbg
		np_thr++;
		vec2pix_ring(n_side,pos,&ipix);
		specin=draw_specin(rng_thr);

		for(inu=0;inu<n_nu;inu++) {
		  double nu=nu_arr[inu];
		  double temperature=t_prefactor*s_nu*
		    pow((1+red_true)*nu/NU0_PSOURCES,-specin)/(nu*nu);
#pragma omp atomic
		  T_maps[ipix+n_pix_ang*inu]+=temperature;
		  //		  T_maps[inu+n_nu*ipix]+=temperature;
		}
	      }
	    }
	  }
	}
      }
    }//end omp for
    
    //dbg 
#pragma omp critical
    {
      np_total+=np_thr;
    }
    
  }//end omp parallel
  
  //dbg
  printf("There are %ld sources over the threshold\n",np_total);

  free(nu_arr);
  timer(2);
  printf("\n");
    
  return T_maps;
}

flouble *mk_T_maps(unsigned int *seed,flouble *mass_grid,flouble *dz_grid)
{
  int ii;
  flouble *T_maps;
  long n_pix_total=nside2npix(n_side)*n_nu;
  double x_sub[N_SUBPART],y_sub[N_SUBPART],z_sub[N_SUBPART];
  double dnu=(nu_max-nu_min)/n_nu;
  //Use this if the mass grid is in units of 2.77459457E11*OmegaB/hhub M_sun
  double M2T_factor=90.057156*OmegaB*hhub*nside2npix(n_side)/(4*M_PI*dnu);

  gsl_rng *rng=init_rng(*seed);
  double lcell=l_box/n_grid;
  for(ii=0;ii<N_SUBPART;ii++) {
    x_sub[ii]=lcell*(rng_01(rng)-0.5);
    y_sub[ii]=lcell*(rng_01(rng)-0.5);
    z_sub[ii]=lcell*(rng_01(rng)-0.5);
  }
  end_rng(rng);

  printf("*** Making maps\n");

  T_maps=(flouble *)calloc(n_pix_total,sizeof(flouble));
  if(T_maps==NULL) error_mem_out();

  timer(0);
#pragma omp parallel default(none)				\
  shared(mass_grid,dz_grid,n_grid,M2T_factor,T_maps,dnu)	\
  shared(l_box,pos_obs,n_side,nu_min,n_nu,x_sub,y_sub,z_sub)
  {
    long n_pix_ang=nside2npix(n_side);
    double dx=l_box/n_grid;
    double inv_dnu=1./dnu;

    int iz;
#pragma omp single 
    {
      printf(" Collecting masses\n");
    }
#pragma omp for nowait
    for(iz=0;iz<n_grid;iz++) {
      double z0=dx*(iz+0.5)-pos_obs[2];
      int iy;
      for(iy=0;iy<n_grid;iy++) {
	double y0=dx*(iy+0.5)-pos_obs[1];
	int ix;
	for(ix=0;ix<n_grid;ix++) {
	  int isub;
	  lint index=ix+n_grid*((lint)(iy+n_grid*iz));
	  double x0=dx*(ix+0.5)-pos_obs[0];
	  double mass_sub=mass_grid[index]/N_SUBPART;
	  //	  double dz_rsd=0;//w.o. RSD
	  double dz_rsd=(double)(dz_grid[index]);//w. RSD
	  for(isub=0;isub<N_SUBPART;isub++) {
	    double x=x0+x_sub[isub];
	    double y=y0+y_sub[isub];
	    double z=z0+z_sub[isub];
	    double r=sqrt(x*x+y*y+z*z);
	    double redshift=z_of_r(r)+dz_rsd;
	    double nu=NU_21/(1+redshift);
	    int inu=(int)(inv_dnu*(nu-nu_min));
	    if((inu>=0)&&(inu<n_nu)) {
	      long ipix;
	      double pos[3]={x,y,z};
	    
	      vec2pix_ring(n_side,pos,&ipix);
#pragma omp atomic
	      T_maps[ipix+n_pix_ang*inu]+=mass_sub;
	    }
	  }
	}
      }
    } //end pragma omp for
#pragma omp barrier

    int inu;
#pragma omp single 
    {
      printf(" Normalizing to temperature\n");
    }
#pragma omp for nowait
    for(inu=0;inu<n_nu;inu++) {
      long ipix;
      double nu=nu_min+(inu+0.5)*dnu;
      double r=r_of_z(NU_21/nu-1);
      double prefac=M2T_factor/(r*r);
      for(ipix=0;ipix<n_pix_ang;ipix++) {
	long index=ipix+n_pix_ang*inu;
	T_maps[index]*=prefac;
      }
    } //end pragma omp for
  } //end pragma omp parallel
  timer(2);
  printf("\n");

  return T_maps;
}
