///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of GetHI.                                       //
//                                                                   //
// GetHI is free software: you can redistribute it and/or modify it  //
// under the terms of the GNU General Public License as published by //
// the Free Software Foundation, either version 3 of the License, or //
// (at your option) any later version.                               //
//                                                                   //
// GetHI is distributed in the hope that it will be useful, but      //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with GetHI.  If not, see <http://www.gnu.org/licenses/>.    //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_rng.h>
#include "common.h"

#define LX_TOY 2.11
#define ALPHA_TOY -0.633
//#define RHO_TOY 3.17E-4
#define RHO_TOY 1.17E-4

#define NZ_PSOURCES 256
#define NL_PSOURCES 256
#define LLOGMIN -5.
#define LLOGMAX 6.

#ifndef SPECIN_VARIANCE
#define SPECIN_VARIANCE 100.0
#endif //_SPECIN_VARIANCE
#define SPECIN_MEAN 1.0

static double nz_psources_arr[NZ_PSOURCES];
static double max_Lpdf_arr[NZ_PSOURCES];
static double glob_inv_dz;
static double glob_dz;

static double schechter(double L)
{
  //////
  // Schechter function
  return RHO_TOY*pow(L/LX_TOY,ALPHA_TOY)*exp(-L/LX_TOY)/LX_TOY;
}

static double l_z_function(double L,double z)
{
  //////
  // Returns the luminosity function for point sources
  // at redshift z with L given in units of 10^22 W/Hz
  //   dn(z,L)/dlog10(L)
  double fz;

  if(z<1.5) fz=pow((1+z),3.1);
  else fz=17.124;

  return fz*schechter(L)*L*M_LN10;
}

double n_of_z_psources(double z)
{
  //////
  // Returns the number density of point sources
  // as a function of redshift integrated over luminosities
  if(z<z_min) return -1;
  else {
    int iz=(int)(z*glob_inv_dz);
    
    if((iz>=NZ_PSOURCES)||(iz<0)) return -1;
    else if(iz==NZ_PSOURCES-1) return nz_psources_arr[NZ_PSOURCES-1];
    else {
      double zi=iz*glob_dz;
      double nz=nz_psources_arr[iz]+
	(nz_psources_arr[iz+1]-nz_psources_arr[iz])*
	(z-zi)*glob_inv_dz;
      return nz;
    }
  }
}

static double l_distribution(double L,double z)
{
  //////
  // Returns the pdf P(log10(L)|z)
  return l_z_function(L,z)/n_of_z_psources(z);
}

void setup_psources(void)
{
  //////
  // Calculates arrays and parameters for the luminosity
  // function of point sources
  int ii;
  double dlogL=(LLOGMAX-LLOGMIN)/NL_PSOURCES;
  glob_dz=z_max/NZ_PSOURCES;
  glob_inv_dz=1./glob_dz;

  for(ii=0;ii<NZ_PSOURCES;ii++) {
    int jj;
    double max_dist;
    double z=ii*glob_dz;
    
    nz_psources_arr[ii]=0;
    max_dist=-1;
    for(jj=0;jj<NL_PSOURCES;jj++) {
      double logL=LLOGMIN+(jj+0.5)*dlogL;
      double lfunc=l_z_function(pow(10.,logL),z);
      
      if(lfunc>=max_dist) max_dist=lfunc;
      nz_psources_arr[ii]+=dlogL*lfunc;
    }
    if(nz_psources_arr[ii]>0)
      max_Lpdf_arr[ii]=1.1*max_dist/nz_psources_arr[ii];
    else
      max_Lpdf_arr[ii]=0;
  }
}

double draw_luminosity(double z,gsl_rng *rng)
{
  //////
  // Generates a random luminosity in units of 10^22 W/Hz
  // following the luminosity function of point sources
  if((z>=z_max)||(z<z_min)) return 0;
  else {
    double u,logL,L;
    int accept=0;
    double maxpdf=max_Lpdf_arr[(int)(z*glob_inv_dz)];
    
    while(!accept) {
      u=rng_01(rng);
      logL=LLOGMIN+(LLOGMAX-LLOGMIN)*rng_01(rng);
      L=pow(10.,logL);
      if(u*maxpdf<l_distribution(L,z))
	accept=1;
    }

    return L;
  }
}

double draw_specin(gsl_rng *rng)
{
  //////
  // Draws a spectral index from a Gaussian distribution
  // of mean SPECIN_MEAN and variance sigma^2=SPECIN_VARIANCE
#ifdef _DEBUG
  return SPECIN_MEAN;
#else //_DEBUG
  double module,phase;

  rng_delta_gauss(&module,&phase,rng,2*SPECIN_VARIANCE);

  return SPECIN_MEAN+module*cos(phase);
#endif //_DEBUG
}

double bias_psources(double z)
{
  return 1;
}
